#include "command_buffers.h"
#include "gpu.h"
#include "swapchain.h"

#include "runtime/renderer_api.h"

#define ONE_SECOND_NS 1000000000u

void RT_RENDERER_API_FN(BeginFrame)(unsigned int frame_id) {
    g_gpu.current_frame_id = frame_id;

    rt_frame_data *frame = rtGetFrameData(frame_id);

    /* Wait until the previous frame is done */
    VkFence fence = g_swapchain.image_fences[frame_id % g_swapchain.image_count];
    RT_VK_CHECK(vkWaitForFences(g_gpu.device, 1, &fence, VK_TRUE, ONE_SECOND_NS));
    RT_VK_CHECK(vkResetFences(g_gpu.device, 1, &fence));

    rtResetCommandPools(frame_id);

    VkResult acquire_res = vkAcquireNextImageKHR(g_gpu.device,
                                                 g_swapchain.swapchain,
                                                 ONE_SECOND_NS,
                                                 frame->image_available,
                                                 fence,
                                                 &frame->swapchain_image_index);
    if (acquire_res == VK_SUBOPTIMAL_KHR || acquire_res == VK_ERROR_OUT_OF_DATE_KHR) {
        /* We need to recreate the swapchain and try again */
        rtLog("vk", "Swapchain has become suboptimal and needs to be re-created.");
        vkDeviceWaitIdle(g_gpu.device);
        if (rtRecreateSwapchain() != RT_SUCCESS) {
            rtReportError("vk", "Failed to recreate the swapchain.");
            return;
        }
        rtRenBeginFrame(frame_id);
    } else if (acquire_res != VK_SUCCESS) {
        rtReportError("vk", "vkAcquireNextImageKHR failed: %u", acquire_res);
    }
}

void RT_RENDERER_API_FN(EndFrame)(unsigned int frame_id) {
    rt_frame_data *frame = rtGetFrameData(frame_id);

    uint32_t image_index = frame->swapchain_image_index;

    VkPresentInfoKHR present_info = {
        .sType              = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR,
        .pImageIndices      = &image_index,
        .pSwapchains        = &g_swapchain.swapchain,
        .swapchainCount     = 1,
        .pWaitSemaphores    = &frame->image_available,
        .waitSemaphoreCount = 1,
    };

    VkResult res = vkQueuePresentKHR(g_gpu.present_queue, &present_info);
    if (res != VK_SUCCESS) {
        rtReportError("vk", "vkQueuePresentKHR failed: %u", res);
    }
}